/*
 * Copyright (C) 2013 Scott McCracken
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 *     Adaptacin basada en el archivo: BluetoothService.java
 *    del proyecto de ejemplo BluetoothChat
 *    The Android Open Source Project
 */
package com.inforbooks.bluetooth;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.UUID;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothServerSocket;
import android.bluetooth.BluetoothSocket;
import android.content.Context;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.util.Log;


public class BluetoothService {
	// Objeto auxiliar para gestionar los subprocesos
	// de aceptacin de la conexin, conexin
	// y gestin de la transferencia de datos
	// durante la conexin

	// Est compuesta por esta clase BluetoothService y tres clases
	// auxiliares derivada de Thread:
	// AceptaConexionThread: 	Escucha conexiones
	// ConexionSocketThread: 	Conecta con un dispositivo y
	//					suspende el descubrimiento
	// TransferenciasThread: Gestiona la conexin y 
	//					la transferencia de datos.

	// Mtodos pblicos:	
	// constructor
	// write()

	// Mtodos pblicos (sincronizados):
	// conectar()
	// conectado()
	// getEstado()
	// start()
	// stop()

	// Mtodos privados:	
	// perdioConexin()
	// ErrorConexion()

	// Mtodos privados (sincronizados):	
	// setestado()

	// Variable de ayuda a la depuracin
	private static final String TAG = "BluetoothService";
	// Si true, crea log de debug
	private static final boolean D = true;

	// Nombre para el registro SDP
	// (Service Discovery Protocol) cuando
	// se crea un conector servidor
	private static final String NAME_SECURE = "BluetoothSeguro";
	private static final String NAME_INSECURE = "BluetoothInseguro";

	// UUID nico para la aplicacin
	private static final UUID MY_UUID_SECURE =
			UUID.fromString("fa87c0d0-afac-11de-8a39-0800200c9a66");
	private static final UUID MY_UUID_INSECURE =
			UUID.fromString("8ce255c0-200a-11e0-ac64-0800200c9a66");

	// Miembros privados de la clase BluetoothService
	private final BluetoothAdapter mAdaptador;
	private final Handler mHandler;

	// Threads
	private AceptaConexionThread mAceptaConexionSeguraThread;
	private AceptaConexionThread mAceptaConexionInseguraThread;
	private ConexionSocketThread mConexionSocketThread;
	private TransferenciasThread mTransferenciasThread;

	// Control del estado de la conexin
	private int mEstado;

	// Constantes que sirven para conocer el
	// estado de la conexin
	// Indica que ahora no se hace nada
	public static final int ESTADO_INACTIVO = 0; 
	// Indica que estamos a la espera de conexiones
	// entrantes
	public static final int ESTADO_ATENTO = 1; 
	// Indica que estamos iniciando una conexin
	// saliente
	public static final int ESTADO_CONECTANDO = 2; 
	// Indica que ahora estamos conectados
	// a un dispositivo remoto
	public static final int ESTADO_CONECTADO = 3;   


	// Constructor del objeto auxiliar
	public BluetoothService(Context context, Handler handler) {
		if (D) Log.d(TAG, "+in: BluetoothService()");

		mAdaptador = BluetoothAdapter.getDefaultAdapter();
		mEstado = ESTADO_INACTIVO;
		mHandler = handler;
	}

	// Asigna la propiedad que mantiene el estado
	// de la conexin
	private synchronized void setEstado(int estado) {
		if (D) Log.d(TAG, "setEstado(): Cambia estado de " + mEstado + " a -> " + estado);

		// Actualiza la propiedad de la clase
		mEstado = estado;

		// Informa el nuevo estado al Handler 
		// para que se pueda actualizar la actividad principal
		mHandler.obtainMessage(MainActivity.MSG_CAMBIO_ESTADO, estado, -1).sendToTarget();
	}

	// Devuelve el estado actual de la conexin
	public synchronized int getEstado() {
		return mEstado;
	}


	// Inicio del servicio
	// Inicia el subproceso (Thread) que hace
	// de servidor
	public synchronized void start() {
		if (D) Log.d(TAG, "+in: start()");

		// Cancela el subproceso que est
		// intentando de realizar una conexin
		if (mConexionSocketThread != null) {mConexionSocketThread.cancel(); mConexionSocketThread = null;}

		// Cancela el subproceso que actualmente
		// sea una conexin vigente 
		if (mTransferenciasThread != null) {mTransferenciasThread.cancel(); mTransferenciasThread = null;}

		// Estado: a la escucha de peticiones
		setEstado(ESTADO_ATENTO);

		// Inicia el subproceso que acta como
		// servidor escuchando la llegada de peticiones     
		if (mAceptaConexionSeguraThread == null) {
			// Seguro
			mAceptaConexionSeguraThread = new AceptaConexionThread(true);
			mAceptaConexionSeguraThread.start();
		}
		if (mAceptaConexionInseguraThread == null) {
			// Inseguro
			mAceptaConexionInseguraThread = new AceptaConexionThread(false);
			mAceptaConexionInseguraThread.start();
		}
	}


	// Inicia la conexin
	// mtodo invocado desde la actividad principal
	// El usuario ya eligi un dispositivo con el quue
	// se conectar y si es un dispositivo emparejado o nuevo.
	// El mtodo es synchronized lo que nos asegura que en un momento
	// slo se ejecuta una invocacin al mtodo y si hubiese otras
	// invocaciones, stas esperarn la finalizacin de la ejecucin
	// vigente. 
	public synchronized void conectar(BluetoothDevice dispositivo, boolean seguro) {
		if (D) Log.d(TAG, "*in: conectar() " + dispositivo + " ,tipo seguro? " + seguro);

		// Cancela subprocesos que estn intentando
		// una conexin
		if (mEstado == ESTADO_CONECTANDO) {
			if (mConexionSocketThread != null) {mConexionSocketThread.cancel(); mConexionSocketThread = null;}
		}

		// Cancela todo subproceso que actualmente
		// est ejecutando una conexin  
		if (mTransferenciasThread != null) {mTransferenciasThread.cancel(); mTransferenciasThread = null;}

		// Inicia el subproceso para conectar con el 
		// dispositivo remoto requerido 
		// Crea una instancia de ConexionSocketThread y 
		// despus invoca al mtodo start()
		mConexionSocketThread = new ConexionSocketThread(dispositivo, seguro);

		// Inicia la ejecucin del subproceso
		mConexionSocketThread.start();

		// Estado: intentando conexin
		setEstado(ESTADO_CONECTANDO);
	}

	// Dispositivo ya conectado
	// Se inicia un subproceso que gestiona la conexin
	// parmetros: 
	// socket: 	el objeto BluetoothSocket 
	// 			al que se ha hecho la conexin
	// device: objeto BluetoothDevice que ha sido conectado
	public synchronized void conectado(BluetoothSocket socket, 
			BluetoothDevice dispositivo, 
			final String socketType) {
		if (D) Log.d(TAG, "*in: conectado() " + dispositivo + ", tipo: " + socketType);

		// Cancela el subproceso que estableci la conexin
		if (mConexionSocketThread != null) {mConexionSocketThread.cancel(); mConexionSocketThread = null;}

		// Cancela cualquier subproceso previo que gestionan 
		// conexiones vigentes;
		// es decir, del mismo tipo del que crearemos ahora
		if (mTransferenciasThread != null) {mTransferenciasThread.cancel(); mTransferenciasThread = null;}

		// Cancela los subproceso de aceptacin de conexin
		if (mAceptaConexionSeguraThread != null) {
			mAceptaConexionSeguraThread.cancel();
			mAceptaConexionSeguraThread = null;
		}
		if (mAceptaConexionInseguraThread != null) {
			mAceptaConexionInseguraThread.cancel();
			mAceptaConexionInseguraThread = null;
		}

		// Inicia el subproceso que gestiona la conexin
		// y que establece las transferencias
		mTransferenciasThread = new TransferenciasThread(socket, socketType);
		mTransferenciasThread.start();

		// Enva el nombre del dispositivo conectado
		// a la interfaz de la actividad        
		Message msg = mHandler.obtainMessage(MainActivity.MSG_NOMBRE_DISPOSITIVO);
		Bundle bundle = new Bundle();
		bundle.putString(MainActivity.NOMBRE_DISPOSITIVO, dispositivo.getName());
		msg.setData(bundle);
		mHandler.sendMessage(msg);

		// Establece el estado conectado        
		setEstado(ESTADO_CONECTADO);
	}

	// Detiene todos los subprocesos   
	public synchronized void stop() {
		if (D) Log.d(TAG, "* in: stop()");

		if (mConexionSocketThread != null) {
			mConexionSocketThread.cancel();
			mConexionSocketThread = null;
		}

		if (mTransferenciasThread != null) {
			mTransferenciasThread.cancel();
			mTransferenciasThread = null;
		}

		if (mAceptaConexionSeguraThread != null) {
			mAceptaConexionSeguraThread.cancel();
			mAceptaConexionSeguraThread = null;
		}

		if (mAceptaConexionInseguraThread != null) {
			mAceptaConexionInseguraThread.cancel();
			mAceptaConexionInseguraThread = null;
		}

		// Actualiza el estado
		// ahora ya no se hace nada
		setEstado(ESTADO_INACTIVO);
	}

	// Escribe en el flujo de datos de salida
	// utilizando el subproceso
	// que gestiona los flujos de entrada
	// y salida
	public void write(byte[] out) {
		if (D) Log.d(TAG, "* in: write()");

		// Crea un objeto TransferenciasThread
		TransferenciasThread r;

		// Sincroniza para obtener
		// el acceso al subproceso
		// que gestiona los flujos de entrada
		// y salida
		synchronized (this) {
			if (mEstado != ESTADO_CONECTADO) return;
			r = mTransferenciasThread;
		}
		// Ejecuta la escritura
		r.write(out);
	}


	// Excepcin al intentar la conexin
	private void errorConexion() {
		if (D) Log.d(TAG, "* in: errorConexion()");

		// Enva el mensaje del error a la actividad principal
		Message msg = mHandler.obtainMessage(MainActivity.MSG_TOAST);
		Bundle bundle = new Bundle();
		bundle.putString(MainActivity.TOAST, "No se puede conectar al dispositivo");
		msg.setData(bundle);
		mHandler.sendMessage(msg);

		// Inicia el servicio para reiniciar
		// en el modo listening  
		BluetoothService.this.start();
	}


	// Se perdi la conexin. 
	private void perdioConexion() {
		if (D) Log.d(TAG, "* in: perdioConexion()");

		// Se enva un mensaje a la actividad 
		Message msg = mHandler.obtainMessage(MainActivity.MSG_TOAST);
		Bundle bundle = new Bundle();
		bundle.putString(MainActivity.TOAST, "Se perdi la conexin con el dispositivo");
		msg.setData(bundle);
		mHandler.sendMessage(msg);

		// reinicia el servicio para 
		// seguir en el modo listening  
		BluetoothService.this.start();
	}

	//******************************************
	// Clases internas
	//******************************************


	//***********************************
	// Clase AceptaConexionThread
	//***********************************

	// Subproceso que se ejecuta mientras se estn esperando
	// conexiones entrantes. Cuando se acepte una conexin
	// el subproceso finaliza (si es que no se cancela antes)
	private class AceptaConexionThread extends Thread {

		// Mtodos (todos pblicos): 
		// constructor
		// run()
		// cancel()

		// Objeto BluetoothServerSocket
		private final BluetoothServerSocket mmServerSocket;
		private String mSocketType;

		// Constructor de la clase
		public AceptaConexionThread(boolean secure) {
			if (D) Log.d(TAG, "* in: constructor AceptaConexionThread");

			BluetoothServerSocket tmp = null;
			// Seguro o inseguro
			mSocketType = secure ? "Secure":"Insecure";

			// Crea un conector servidor
			// que atiende las peticiones de conexin           
			try {
				if (secure) {
					// (un cliente remoto que se conecte por 
					// este socket se autenticar y la 
					// comunicacin ser cifrada)
					tmp = mAdaptador.listenUsingRfcommWithServiceRecord(NAME_SECURE,
							MY_UUID_SECURE);
				} else {
					tmp = mAdaptador.listenUsingInsecureRfcommWithServiceRecord(
							NAME_INSECURE, MY_UUID_INSECURE);
				}
			} catch (IOException e) {
				Log.e(TAG, "Fall la escucha del tipo de conector: " + mSocketType, e);
			}
			// Se asigna el objeto BluetoothServerSocket
			// con el objeto recibido del adaptador
			mmServerSocket = tmp;
		}

		// Ejecucin (al invocar al mtodo start())
		public void run() {
			// Tipo de conector: seguro o inseguro
			if (D) Log.d(TAG, "*in: run() AceptaConexionThread: Tipo: " + mSocketType);

			setName("AceptaConexionThread" + mSocketType);

			BluetoothSocket socket = null;

			// Mientras no est conectado, se sigue el bucle
			// BUCLE
			while (mEstado != ESTADO_CONECTADO) {
				try {
					// Este mtodo bloquea y slo retornar
					// con una conexin o una excepcin
					// El bloqueo se puede interrumpir desde afuera
					// utilizando el mtodo close(); por eso se
					// incluye despus el mtodo cancel.
					socket = mmServerSocket.accept();
				} catch (IOException e) {
					Log.e(TAG, "Mtodo accept() gener una excepcin", e);
					break;
				}

				//
				// Conexin aceptada
				if (socket != null) {
					synchronized (BluetoothService.this) {
						// Comprueba el estado
						switch (mEstado) {
						case ESTADO_ATENTO:
						case ESTADO_CONECTANDO:
							// Inicia el subproceso
							conectado(socket, socket.getRemoteDevice(),
									mSocketType);
							break;
						case ESTADO_INACTIVO:
						case ESTADO_CONECTADO:
							// Ya conectado, se finaliza el conector.
							try {
								socket.close();
							} catch (IOException e) {
								Log.e(TAG, "No se pudo cerrar una conexin", e);
							}
							break;
						}
					}
				}
			}
			if (D) Log.i(TAG, "FIN AceptaConexionThread, Tipo: " + mSocketType);

		}

		// Para poder cancelar el bloqueo de accept()
		// desde otro subproceso
		public void cancel() {
			if (D) Log.d(TAG, "* in: cancel(): Se cancela el conector " + this);
			try {
				mmServerSocket.close();
			} catch (IOException e) {
				Log.e(TAG, "Excepcin al cerrar el servidor.", e);
			}
		}
	}

	//***********************************
	// Clase ConexionSocketThread
	//***********************************


	// Clase para gestionar el subproceso de conexin
	private class ConexionSocketThread extends Thread {

		// Mtodos (todos pblicos):
		// constructor
		// run()
		// cancel()

		private final BluetoothSocket mmSocket;
		private final BluetoothDevice mmDevice;
		private String mSocketType;

		// Constructor
		public ConexionSocketThread(BluetoothDevice dispositivo, boolean seguro) {
			if (D) Log.d(TAG, "* constructor: ConexionSocketThread");

			mmDevice = dispositivo;
			BluetoothSocket tmp = null;
			mSocketType = seguro ? "Secure" : "Insecure";

			// Obtiene un BluetoothSocket para una conexin
			// con un objeto BluetoothDevice
			try {
				if (seguro) {
					// crea un BluetoothSocket
					// conexin RFCOMM  segura
					tmp = dispositivo.createRfcommSocketToServiceRecord(
							MY_UUID_SECURE);
				} else {
					// crea un BluetoothSocket
					// conexin RFCOMM  insegura
					tmp = dispositivo.createInsecureRfcommSocketToServiceRecord(
							MY_UUID_INSECURE);
				}
			} catch (IOException e) {
				Log.e(TAG, "Fall la creacin del conector de tipo: " + mSocketType, e);
			}

			// Objeto BluetoothSocket
			mmSocket = tmp;
		}

		// Ejecuta el subproceso
		public void run() {
			if (D) Log.d(TAG, "* run -> ConexionSocketThread");

			setName("ConexionSocketThread" + mSocketType);

			// Se cancela el proceso de descubrimiento
			// porque hace que la conexin sea ms lenta
			mAdaptador.cancelDiscovery();

			// realiza la conexin con el objeto 
			// BluetoothSocket
			try {
				// El mtodo connect bloquea
				// y  terminar cuando se haga la conexin
				// o con una excepcin  
				mmSocket.connect();
			} catch (IOException e) {
				// Cierra el conector
				try {
					// Cierra el conector
					mmSocket.close();
				} catch (IOException e2) {
					Log.e(TAG, "Excepcin de E/S al intentar cerrar el conector " + mSocketType, e2);
				}
				// Fall la conexin
				errorConexion();
				return;
			}

			// Este paso se hace sincronizado.
			// La conexin ya est realizada
			// por lo que el subproceso de ConexionSocketThread
			// ya no es necesario
			synchronized (BluetoothService.this) {

				mConexionSocketThread = null;
			}

			// Inicia el subproceso 
			// que gestiona la conexin ya conectada  
			conectado(mmSocket, mmDevice, mSocketType);
		}

		public void cancel() {
			if (D) Log.d(TAG, "* in: cancel() ConexionSocketThread");

			try {
				mmSocket.close();
			} catch (IOException e) {
				Log.e(TAG, "Excepcin en el cierre de la conexin " + mSocketType, e);
			}
		}
	}


	//***********************************
	// Clase TransferenciasThread
	//***********************************

	// Esta clase se ejecuta en un subproceso 
	// durante la conexin con un dispositivo remoto
	// gestiona las transferencias entrantes y salientes
	private class TransferenciasThread extends Thread {

		// Mtodos (todos pblicos):
		// constructor
		// run()
		// write()
		// cancel()

		private final BluetoothSocket mmSocket;

		// Crea los flujos de datos
		private final InputStream mmInStream;
		private final OutputStream mmOutStream;

		// Constructor 
		// parmetros: 
		// Un conector BluetoothSocket para gestionar la conexin 
		// Un tipo de conector (seguro/inseguro)
		public TransferenciasThread(BluetoothSocket socket, String socketType) {
			Log.d(TAG, "* in: constructor TransferenciasThread: " + socketType);

			mmSocket = socket;
			InputStream tmpIn = null;
			OutputStream tmpOut = null;

			// Obtiene los objetos de flujos de datos
			// del conector
			try {
				tmpIn = socket.getInputStream();
				tmpOut = socket.getOutputStream();
			} catch (IOException e) {
				Log.e(TAG, "No se cre el conector temporal", e);
			}

			mmInStream = tmpIn;
			mmOutStream = tmpOut;
		}

		// Se ejecuta el subproceso
		public void run() {
			if (D) Log.d(TAG, "* run: TransferenciasThread");

			byte[] buffer = new byte[1024];
			int bytes;

			// Se mantiene la lectura mientras
			// se mantenga la conexin
			while (true) {
				try {
					// Lee desde el objeto InputStream
					// en bytes est la cantidad
					// de los bytes ledos
					bytes = mmInStream.read(buffer);

					// Enva los bytes obtenidos a la interfaz
					// de la actividad
					mHandler.obtainMessage(MainActivity.MSG_LEER, bytes, -1, buffer)
					.sendToTarget();
				} catch (IOException e) {
					Log.e(TAG, "Desconectado", e);
					// Se perdi la conexin
					perdioConexion();
					// Inicia el servicio para seguir en el 
					// modo listening  
					BluetoothService.this.start();
					break;
				}
			}
		}


		// Mtodo para escribir en el flujo de datos
		// de salida
		// Parmetros:
		// bytes que se transmiten
		public void write(byte[] buffer) {
			if (D) Log.d(TAG, "* write(): TransferenciasThread");

			try {
				mmOutStream.write(buffer);

				// Comparte el mensaje enviado
				// hacia la interfaz de la
				// actividad principal 
				mHandler.obtainMessage(MainActivity.MSG_ESCRIBIR, -1, -1, buffer)
				.sendToTarget();

			} catch (IOException e) {
				Log.e(TAG, "Excepcin de E/S en la grabacin del flujo de datos", e);
			}
		}

		public void cancel() {
			if (D) Log.d(TAG, "* cancel(): TransferenciasThread");

			try {
				mmSocket.close();
			} catch (IOException e) {
				Log.e(TAG, "Excepcin de E/S al cerrar el conector", e);
			}
		}
	}
}
